/*
 * HammockClassLoader.java
 *  
 * Copyright 2009 C.A. Meijer
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.hammingweight.hammock.eclipse.hammockmaker.classloader;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;

import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.jdt.core.IClasspathEntry;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.JavaModelException;

public class HammockClassLoader extends ClassLoader {

	private IJavaProject project;
	
	private byte[] readInputStream(InputStream is) throws IOException {
		byte[] buf = new byte[128];
		int len = 0;
		
		int numRead;
		while ((numRead = is.read(buf, len, buf.length - len)) > 0) {
			len += numRead;
			if (len == buf.length) {
				byte[] buf2 = new byte[2 * len];
				System.arraycopy(buf, 0, buf2, 0, len);
				buf = buf2;
			}
		}

		byte[] buf2 = new byte[len];
		System.arraycopy(buf, 0, buf2, 0, len);
		return buf2;
	}
	
	private Class<?> readFromOutputLocation(String name) {
		String path = ResourcesPlugin.getWorkspace().getRoot().getLocation().toString();
		if (!path.endsWith("/")) {
			path += "/";
		}
		try {
			path += project.getOutputLocation().toString();
			if (!path.endsWith("/")) {
				path += "/";
			}
			path += name.replace('.', '/') + ".class";
			File file = new File(path);
			FileInputStream fis = new FileInputStream(file);
			byte[] buf = readInputStream(fis);
			return super.defineClass(name, buf, 0, buf.length);
		} catch (JavaModelException e) {
			throw new RuntimeException("foo");
		} catch (FileNotFoundException e) {
			return null;
		} catch (IOException e) {
			throw new RuntimeException(e.getMessage());
		}
	}
	
	private Class<?> readFromJar(String path, String name) {
		try {
			String pName = name.replace('.', '/') + ".class";
			JarFile jarFile = new JarFile(path);
			ZipEntry ze = jarFile.getEntry(pName);
			if (ze == null) {
				return null;
			}
			InputStream is = jarFile.getInputStream(ze);
			byte[] buf = readInputStream(is);
			return defineClass(name, buf, 0, buf.length);
		} catch (FileNotFoundException e) {
			return null;
		} catch (IOException e) {
			return null;
		}
	}
	
	private Class<?> readFromJar(IClasspathEntry entry, String name) {
		Class<?> clazz = readFromJar(entry.getPath().toString(), name);
		if (clazz != null) {
			return clazz;
		}
		String root = ResourcesPlugin.getWorkspace().getRoot().getLocation().toString();
		//throw new RuntimeException(root + "/" + entry.getPath().toString());
		return readFromJar(root + entry.getPath().toString(), name);
	}
	
	private Class<?> readLibraries(String name) {
		IClasspathEntry[] entries;
		try {
			entries = this.project.getResolvedClasspath(false);
		} catch (JavaModelException e) {
			throw new RuntimeException("foobar");
		}
		String s = "";
		for (IClasspathEntry entry : entries) {
			if (entry.getEntryKind() != IClasspathEntry.CPE_LIBRARY) {
				continue;
			}
			s += entry.getPath() + "\n";
			String root = ResourcesPlugin.getWorkspace().getRoot().getLocation().toString();
			s += root + entry.getPath().toString().substring(1) + "\n";
			Class<?> clazz = this.readFromJar(entry, name);
			if (clazz != null) {
				return clazz;
			}
		}
		throw new RuntimeException(s);
	}
	
	public HammockClassLoader(IJavaProject project) {
		this.project = project;
	}
	
	protected Class<?> findClass(String name) throws ClassNotFoundException {
		Class<?> clazz = readFromOutputLocation(name);
		if (clazz == null) {
			clazz = readLibraries(name);
		}
		return clazz;
		//return this.getClass().getClassLoader().loadClass(name);
	}
}
