/*
 * HammockHandler.java
 *  
 * Copyright 2009 C.A. Meijer
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.hammingweight.hammock.eclipse.hammockmaker.handlers;

import java.io.File;

import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.jdt.core.IClassFile;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.ui.handlers.HandlerUtil;

import com.hammingweight.hammock.eclipse.hammockmaker.classloader.HammockClassLoader;
import com.hammingweight.hammock.eclipse.hammockmaker.dialogs.HammockDialogs;
import com.hammingweight.hammock.eclipse.hammockmaker.dialogs.HammockMakerDialog;
import com.hammingweight.hammockmaker.Maker;

/**
 * The HammockHandler class creates Hammock test doubles in response to an
 * Eclipse command.
 */
public class HammockHandler extends AbstractHandler {

	private String directory = null;

	private String packageName = "";

	private boolean isCldc11 = true;

	private boolean isValidDirectory(String directory, boolean dispError) {
		File dir = new File(directory);
		if (dir.exists() && dir.isDirectory()) {
			return true;
		}

		if (dispError) {
			HammockDialogs
					.displayErrorMessage(
							"Cannot create test double.",
							"The output directory is missing or invalid.\nPlease supply a valid output path.");
		}
		return false;
	}

	/**
	 * the command has been executed, so extract the needed information from the
	 * application context:
	 * 
	 * Check that the selected item is a Java source or class file. Prompt the
	 * user for the details of the test double that needs to be created. Create
	 * the test double.
	 */
	public Object execute(ExecutionEvent event) throws ExecutionException {
		ISelection selection = HandlerUtil.getCurrentSelection(event);
		if (!(selection instanceof IStructuredSelection)) {
			String reason = "HammockMaker can't determine the class to mock from "
				+ selection;
			HammockDialogs.displayErrorMessage("Can't create test double", reason);
			throw new ExecutionException(reason);
		}
		IStructuredSelection structuredSelection = (IStructuredSelection) selection;
		Object element = structuredSelection.getFirstElement();
		IType type = null;
		if (element instanceof IClassFile) {
			type = ((IClassFile) element).getType();
		} else if (element instanceof IType) {
			type = (IType) element;
		} else if (element instanceof ICompilationUnit) {
			ICompilationUnit icu = (ICompilationUnit) element;
			try {
				IType[] types = icu.getAllTypes();
				if (types.length == 0) {
					String reason = "HammockMaker can't mock "
						+ element + "\nCan't find class type.";
					HammockDialogs.displayErrorMessage("Can't create test double", reason);
					throw new ExecutionException(reason);
				}
				type = types[0];
			} catch (JavaModelException e) {
				throw new ExecutionException(e.getMessage());
			}
		} else {
			HammockDialogs.displayErrorMessage("Can't create test double", "An unexpected error occurred.");
			throw new ExecutionException("HammockMaker can't mock " + element);
		}

		if (this.directory == null) {
			if (element instanceof IJavaElement) {
				IJavaElement javaElement = (IJavaElement) element;
				this.directory = ResourcesPlugin.getWorkspace().getRoot()
						.getLocation().toString();
				this.directory += javaElement.getJavaProject().getPath()
						.toString();
				if (isValidDirectory(this.directory + "/test", false)) {
					this.directory += "/test";
				} else if (isValidDirectory(this.directory + "/src", false)) {
					this.directory += "/src";
				}
			} else {
				this.directory = "";
			}
		}
		String name = type.getFullyQualifiedName();
		HammockMakerDialog dialog = new HammockMakerDialog(HandlerUtil
				.getActiveShell(event), name, this.packageName, this.directory,
				this.isCldc11);
		while (true) {
			if (dialog.open() == Dialog.OK) {
				this.directory = dialog.getDirectory();
				this.isCldc11 = dialog.isCldc11();
				this.packageName = dialog.getPackage();
				if (this.isValidDirectory(this.directory, true)) {
					if (!this.packageName.equals("")) {
						break;
					}
					if (HammockDialogs
							.displayConfirm("Do you really want to create the mock in the default package?")) {
						break;
					}
				}
				dialog = new HammockMakerDialog(HandlerUtil
						.getActiveShell(event), name, this.packageName,
						this.directory, this.isCldc11);
			} else {
				return null;
			}
		}

		try {
			IJavaProject project = type.getJavaProject();
			ClassLoader cl = new HammockClassLoader(project);
			Class<?> clazz = cl.loadClass(name);
			String clazzName = name.substring(name.lastIndexOf('.') + 1);
			Maker.createMock(clazz, dialog.getPackage(), "Mock" + clazzName,
					dialog.isCldc11(), dialog.getDirectory());
			project.getResource().refreshLocal(IResource.DEPTH_INFINITE, null);
			HammockDialogs
					.displayOk("Test double created.\n\n"
							+ "You may need to refresh your project to see the changes.");
		} catch (Exception e) {
			HammockDialogs.displayErrorMessage("Cannot create test double.", e
					.toString());
		}
		return null;
	}
}
